// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

/*
Package otlptrace contains abstractions for OTLP span exporters.
See the official OTLP span exporter implementations:
  - [go.opentelemetry.io/otel/exporters/otlp/otlptrace/otlptracegrpc],
  - [go.opentelemetry.io/otel/exporters/otlp/otlptrace/otlptracehttp].
*/
package otlptrace // import "go.opentelemetry.io/otel/exporters/otlp/otlptrace"
