// Code generated by go-swagger; DO NOT EDIT.

package container

// This file was generated by the swagger tool.
// Editing this file might prove futile when you re-run the swagger generate command

// WaitExitError container waiting error, if any
//
// swagger:model WaitExitError
type WaitExitError struct {

	// Details of an error
	Message string `json:"Message,omitempty"`
}
