// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Package exemplar provides an implementation of the OpenTelemetry exemplar
// reservoir to be used in metric collection pipelines.
package exemplar // import "go.opentelemetry.io/otel/sdk/metric/exemplar"
